using System;
using System.Diagnostics;
using System.IO.Ports;
using System.Xml;
using Bespoke.Robotics.Common;

namespace Bespoke.Robotics.BrainStem.Configuration
{
    /// <summary>
	/// The SerialPort Xml element.
    /// </summary>
    public class SerialPortElement
    {
        /// <summary>
		/// The name of the element.
        /// </summary>
        public const string ELEMENT_NAME = "serialPort";

        /// <summary>
        /// Gets the serial port.
        /// </summary>
		/// <value>The serial port.</value>
		public string Port
		{
			get
			{
				return mPort;
			}
		}

        /// <summary>
        /// Gets the baud rate.
        /// </summary>
		/// <value>The baud rate.</value>
		public int BaudRate
		{
			get
			{
				return mBaudRate;
			}
		}

        /// <summary>
        /// Gets the number of data bits.
        /// </summary>
		/// <value>The number of data bits.</value>
        public int DataBits
        {
            get
            {
                return mDataBits;
            }
        }

        /// <summary>
        /// Gets the parity.
        /// </summary>
        /// <value>The parity.</value>
        public Parity Parity
        {
            get
            {
                return mParity;
            }
        }

        /// <summary>
        /// Gets the number of stop bits.
        /// </summary>
		/// <value>The number of stop bits.</value>
        public StopBits StopBits
        {
            get
            {
                return mStopBits;
            }
        }

        /// <summary>
		/// Gets the number of milliseconds before a timeout occurs when a read operation does not finish.
        /// </summary>
		/// <value>The number of milliseconds before a timeout occurs when a read operation does not finish.</value>
        public string ReadTimeout
        {
            get
            {
                return mReadTimeout;
            }
        }

        /// <summary>
		/// Initializes a new instance of the SerialPortElement class.
        /// </summary>
		/// <param name="configNode">An Xml node containing the BrainStem element.</param>
        public SerialPortElement(XmlNode configNode)
		{
			Clear();
			ParseConfig(configNode);
		}

        /// <summary>
		/// Reset all configuration data.
        /// </summary>
		public void Clear()
		{
			mPort = String.Empty;
			mBaudRate = 9600;
            mDataBits = 8;
            mParity = Parity.None;
            mStopBits = StopBits.One;
            mReadTimeout = String.Empty;
		}

        /// <summary>
		/// Parse the configuration file.
        /// </summary>
		/// <param name="configNode">An Xml node containing the BrainStem element.</param>
		/// <exception cref="Exception">Thrown if there are any errors reading the BrainStem element.</exception>
		private void ParseConfig(XmlNode configNode)
		{
			try
			{
                mPort = configNode.Attributes[PORT_ATTRIBUTE].Value.Trim();
                mBaudRate = int.Parse(configNode.Attributes[BAUDRATE_ATTRIBUTE].Value.Trim());
                mDataBits = int.Parse(configNode.Attributes[DATABITS_ATTRIBUTE].Value.Trim());
                mParity = (Parity)Enum.Parse(typeof(Parity), configNode.Attributes[PARITY_ATTRIBUTE].Value.Trim());
                mStopBits = (StopBits)Enum.Parse(typeof(StopBits), configNode.Attributes[STOPBITS_ATTRIBUTE].Value.Trim());
                mReadTimeout = configNode.Attributes[READTIMEOUT_ATTRIBUTE].Value.Trim();

                if ((mPort == String.Empty) && (mReadTimeout == String.Empty))
				{
					throw new Exception("Malformed serialPort element.");
				}
			}
			catch (Exception ex)
			{
				Clear();
				Trace.WriteLine(ex.Message);
                throw ex;
			}
		}

		private const string PORT_ATTRIBUTE = "port";
		private const string BAUDRATE_ATTRIBUTE = "baudRate";
        private const string DATABITS_ATTRIBUTE = "dataBits";
        private const string PARITY_ATTRIBUTE = "parity";
        private const string STOPBITS_ATTRIBUTE = "stopBits";
        private const string READTIMEOUT_ATTRIBUTE = "readTimeout";

		private string mPort;
		private int mBaudRate;
        private int mDataBits;
        private Parity mParity;
        private StopBits mStopBits;
        private string mReadTimeout;
    }
}
