using System;
using System.Diagnostics;
using System.Threading;
using System.IO.Ports;
using Bespoke.Robotics.BrainStem.Configuration;
using Bespoke.Robotics.BrainStem.Exceptions;

namespace Bespoke.Robotics.BrainStem
{
    /// <summary>
    /// Represents the method that will handle the PacketReceived event of a Stream object. 
    /// </summary>
    /// <param name="sender">The sender of the event.</param>
    /// <param name="e">An EventArgs object that contains the event data.</param>
    public delegate void PacketReceivedEventHandler(object sender, PacketEventsArgs e);

    /// <summary>
    /// Represents the method that will handle the HeartbeatReceived event of a Stream object. 
    /// </summary>
    /// <param name="sender">The sender of the event.</param>
    /// <param name="e">An EventArgs object that contains the event data.</param>
    public delegate void HeartbeatReceivedEventHandler(object sender, HeartbeatEventsArgs e);

    /// <summary>
    /// Represents a connection to an Acroname BrainStem module.
    /// </summary>
    public sealed class Stream : IDisposable
    {
        /// <summary>
        /// Occurs when a heartbeat packet is received.
        /// </summary>
        public event HeartbeatReceivedEventHandler HeartbeatReceived;

        /// <summary>
        /// Occurs when a packet is received.
        /// </summary>
        /// <remarks>Heartbeat packets are not processed by this event.</remarks>
        public event PacketReceivedEventHandler PacketReceived;

        /// <summary>
        /// Gets the open or closed status of the serial port connection.
        /// </summary>
        public bool IsOpen
        {
            get
            {
                if (mSerialPort == null)
                {
                    return false;
                }
                else
                {
                    return mSerialPort.IsOpen;
                }
            }
        }

        /// <summary>
		/// Initializes a new instance of the Sream class from a configuration file.
        /// </summary>
        public Stream()
        {
            SerialPortElement serialPortConfig = Configuration.Configuration.SerialPort;
            mSerialPort = new SerialPort(serialPortConfig.Port, serialPortConfig.BaudRate, serialPortConfig.Parity,
                serialPortConfig.DataBits, serialPortConfig.StopBits);
            mSerialPort.DataReceived += new SerialDataReceivedEventHandler(mSerialPort_DataReceived);
        }

        /// <summary>
		/// Initializes a new instance of the Stream class.
        /// </summary>
        /// <param name="port">A string indicating the port to be used, for example, COM1</param>
        /// <param name="baudRate">An integer indicating the baud rate. </param>
        /// <param name="parity">A value from the <see cref="Parity"/> enumeration.</param>
        /// <param name="dataBits">An integer indicating the data bits value.</param>
        /// <param name="stopBits">A value from the <see cref="StopBits"/> enumeration.</param>
        public Stream(string port, int baudRate, Parity parity, int dataBits, StopBits stopBits)
        {
            mSerialPort = new SerialPort(port, baudRate, parity, dataBits, stopBits);
            mSerialPort.DataReceived += new SerialDataReceivedEventHandler(mSerialPort_DataReceived);
        }

        /// <summary>
        /// Closes the serial port connection.
        /// </summary>
        public void Close()
        {
            if (mSerialPort.IsOpen)
            {
                mSerialPort.Close();
            }
        }

        /// <summary>
        /// Clean up any resources being used.
        /// </summary>
        public void Dispose()
        {
            if (mSerialPort != null)
            {
                mSerialPort.Dispose();
            }
        }

        /// <summary>
        /// Opens the serial port connection.
        /// </summary>
        public void Open()
        {
            if (mSerialPort.IsOpen)
            {
                throw new StreamAlreadyOpenException();
            }

            mSerialPort.Open();
        }

        /// <summary>
        /// Write a packet to the open stream.
        /// </summary>
        /// <param name="packet">The packet to write.</param>
        /// <exception cref="StreamClosedException">Thrown if the stream is not open.</exception>
        /// <exception cref="PacketDirectionException">Thrown if the specified packet's direction is not
        /// Packet.DirectionType.HostToStem.</exception>
        public void WritePacket(Packet packet)
        {
            if (mSerialPort.IsOpen == false)
            {
                throw new StreamClosedException();
            }

            if (packet.Direction != Packet.DirectionType.HostToStem)
            {
				throw new PacketDirectionException("Outgoing packets must be marked with Packet.DirectionType.HostToStem.");
            }

            byte[] buffer = packet.ToByteArray();
            mSerialPort.Write(buffer, 0, buffer.Length);
        }

        /// <summary>
        /// Raise the HeartbeatReceived event.
        /// </summary>
        /// <param name="e">An EventArgs object that contains the event data.</param>
        private void OnHeartbeatReceived(HeartbeatEventsArgs e)
        {
            if (HeartbeatReceived != null)
            {
                HeartbeatReceived(this, e);
            }
        }

        /// <summary>
        /// Raise the PacketReceived event.
        /// </summary>
        /// <param name="e">An EventArgs object that contains the event data.</param>
        private void OnPacketReceived(PacketEventsArgs e)
        {
            if (PacketReceived != null)
            {
                PacketReceived(this, e);
            }
        }

        /// <summary>
        /// Process data received from the serial port.
        /// </summary>
        /// <param name="sender">The sender of the event.</param>
        /// <param name="e">An EventArgs object that contains the event data.</param>
        private void mSerialPort_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
			Packet packet = null;
			byte iicAddress;
			int length;
			byte[] data = null;

            iicAddress = (byte)mSerialPort.ReadByte();
            length = ReadByte() - 1;
			Packet.CommandCodeType commandCode = (Packet.CommandCodeType)ReadByte();

			if (length > 0)
            {
				data = ReadBytes(length);
            }

			switch (commandCode)
			{
				case Packet.CommandCodeType.HeartBeat:
					if (data != null)
					{
						packet = new Packet(iicAddress, Packet.CommandCodeType.HeartBeat, Packet.DirectionType.StemToHost, data);
						OnHeartbeatReceived(new HeartbeatEventsArgs(packet, (HeartbeatType)data[0]));
					}
					break;

				default:
					packet = new Packet(iicAddress, commandCode, Packet.DirectionType.StemToHost, data);
					OnPacketReceived(new PacketEventsArgs(packet));
					break;
			}
        }

		/// <summary>
		/// Read a single byte from the serial port.
		/// </summary>
		/// <returns>A single byte from the serial port.</returns>
		/// <exception cref="CommandReadTimeoutException">Thrown if the command read timeout expires.</exception>
		/// <remarks>This method waits for the byte to appear before attempting to read it.</remarks>
		private byte ReadByte()
		{
			if (mSerialPort.BytesToRead < 1)
			{
				WaitForBytes(1);
			}

			return (byte)mSerialPort.ReadByte();
		}

		/// <summary>
		/// Read an array of bytes from the serial port.
		/// </summary>
		/// <param name="bytesToRead">The number of bytes to read.</param>
		/// <returns>An array of bytes from the serial port.</returns>
		/// <exception cref="CommandReadTimeoutException">Thrown if the command read timeout expires.</exception>
		/// <remarks>This method waits for the bytes to appear before attempting to read them.</remarks>
		private byte[] ReadBytes(int bytesToRead)
		{
			if (mSerialPort.BytesToRead < bytesToRead)
			{
				WaitForBytes(bytesToRead);
			}

			byte[] buffer = new byte[bytesToRead];
			mSerialPort.Read(buffer, 0, bytesToRead);
			return buffer;
		}

		/// <summary>
		/// Wait for a number of bytes to appear on the serial port.
		/// </summary>
		/// <param name="bytesToRead">The number of bytes to wait for.</param>
		/// <exception cref="CommandReadTimeoutException">Thrown if the command read timeout expires.</exception>
		private void WaitForBytes(int bytesToRead)
		{
			int waitTime = 0;
			while (mSerialPort.BytesToRead < bytesToRead)
			{
				if (waitTime > COMMAND_READ_TIMEOUT)
				{
					throw new CommandReadTimeoutException();
				}

				waitTime += WAIT_SLEEP_TIME;
				Thread.Sleep(WAIT_SLEEP_TIME);
			}
		}

		private const int WAIT_SLEEP_TIME = 10;
		private const int COMMAND_READ_TIMEOUT = 500;

        private SerialPort mSerialPort;
    }
}
