using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading;

using Bespoke.Robotics.Common;
using Bespoke.Robotics.BrainStem;
using Bespoke.Robotics.BrainStem.Configuration;

namespace Bespoke.Robotics.BrainStem.Example
{
    /// <summary>
    /// Main UI for the BrainStem.NET example.
    /// </summary>
    public partial class MainForm : Form
    {
        /// <summary>
        /// Initializes a new instance of the MainForm class.
        /// </summary>
        public MainForm()
        {
            InitializeComponent();

            try
            {
                // Read the BrainStem module address from the Bespoke.Robotics.BrainStem.config file.
                mModuleAddress = Configuration.Configuration.BrainStem.ModuleAddress;

                // Create and open a new Stream object. For initializing the BrainStem.NET CommandManager.
                mStream = new Stream();
                //mStream.HeartbeatReceived += new HeartbeatReceivedEventHandler(mStream_HeartbeatReceived);
                mStream.Open();

                // This is the last call required before sending commands to the BrainStem.
                CommandManager.Init(mStream, Configuration.Configuration.BrainStem.CommandReplyTimeout);

                CommandManager.SoftwareReset(mModuleAddress);
                Thread.Sleep(1000);

                byte servoPosition = CommandManager.GetServoPosition(mModuleAddress, ServoIdType.Servo0);
                SetOutputText("Servo Position: " + servoPosition.ToString() + Environment.NewLine);

                int analogValue = CommandManager.ReadA2DInput(mModuleAddress, AnalogToDigitalIoPinType.Pin0, true, false);
                SetOutputText("Analog Value: " + analogValue.ToString() + Environment.NewLine);

            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.Message);
            }
        }

        /// <summary>
        /// Handle the BrainStem Heartbeat event.
        /// </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">An <see "EventArgs"> object containing the event data.</param>
        private void mStream_HeartbeatReceived(object sender, HeartbeatEventsArgs e)
        {
            bool heartbeatUp = (e.HeartbeatType == HeartbeatType.StemToHostUp);
            SetHeartBeatText(heartbeatUp);
        }

        /// <summary>
        /// 
        /// </summary>
        private void SetOutputText(string output)
        {
            try
            {
                if (mOutputTextBox.InvokeRequired)
                {
                    SetOutputTextHandler handler = new SetOutputTextHandler(SetOutputText);
                    BeginInvoke(handler, new object[] { output });
                }
                else
                {
                    mOutputTextBox.Text += output;
                    mOutputTextBox.SelectionStart = mOutputTextBox.Text.Length;
                    mOutputTextBox.ScrollToCaret();
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
            }
        }

        /// <summary>
        /// Update the heart beat UI.
        /// </summary>
        /// <param name="up">The type of heart beat to set (either up or down).</param>
        private void SetHeartBeatText(bool up)
        {
            try
            {
                if (mOutputTextBox.InvokeRequired)
                {
                    SetHeartbeatButtonHandler handler = new SetHeartbeatButtonHandler(SetHeartBeatText);
                    BeginInvoke(handler, new object[] { up });
                }
                else
                {
                    mOutputTextBox.Text += string.Format("Heartbeat: {0}{1}", (up ? "Up" : "Down"), Environment.NewLine);
                    mOutputTextBox.SelectionStart = mOutputTextBox.Text.Length;
                    mOutputTextBox.ScrollToCaret();
                }
            }
            catch (Exception ex)
            {
                Debug.WriteLine(ex.Message);
            }
        }

        private delegate void SetHeartbeatButtonHandler(bool up);
        private delegate void SetOutputTextHandler(string output);

        private byte mModuleAddress;
        private Stream mStream;
    }
}